#!/usr/bin/env python3

import logging

import click

from .build_csv import _is_reg_row, clean_row, get_raw_rows
from .constants import MODULES, MOST_RECENT_YEAR, OLDEST_YEAR

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def get_mod_table_headers(mod, year):
    previous_row = False
    headers = []
    raw_rows = get_raw_rows(mod, year)

    for page in raw_rows:
        for row in raw_rows[page]:
            row = clean_row(row)
            if len(row) > 2 and _is_reg_row(row) and previous_row not in headers:
                headers.append(previous_row)
            previous_row = row
    return headers


@click.option(
    "--year",
    default=MOST_RECENT_YEAR,
    show_default=True,
    type=click.IntRange(OLDEST_YEAR, MOST_RECENT_YEAR),
    help="Operate on a specific year's folder, "
    f"can be between {OLDEST_YEAR} and {MOST_RECENT_YEAR}",
)
@click.command()
def main(year):
    """Display a list of all the different fields headers found in the four modules.

    Used to define the modules headers hard-coded at the beginning of ./build_csv.py .
    """
    for mod in MODULES:
        headers = get_mod_table_headers(mod, year)
        logger.info(f"\n{mod.upper()}'s headers :")
        for header in headers:
            logger.info(header)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright 2017-2018 Akretion LTDA (http://www.akretion.com)
# @author Raphael Valyi (rvalyi@akretion.com.br)
# License AGPL-3.0 or later (https://www.gnu.org/licenses/agpl).

# ==================================
#
# WORK IN PROGRESS...
#
# ===================================

"""
Synopsis:
    Generate Odoo SPED model definitions.
Usage:
    python generate.py [options]
Options:
    -f, --force
            Overwrite models.py and forms.py without asking.
    --no-class-suffixes
            Do not add suffix "_model" and _form" to generated class names.
    -h, --help
            Show this help message.
"""


import csv
import getopt
import os
import re
import sys
from os import walk

from unidecode import unidecode

# from sped.escrituracao import Escrituracao


PRINT_HIERARCHY = False


class ProgramOptions(object):
    def get_force_(self):
        return self.force_

    def set_force_(self, force):
        self.force_ = force

    force = property(get_force_, set_force_)


class Writer(object):
    def __init__(self, outfilename, stdout_also=False, mode="w"):
        self.outfilename = outfilename
        self.outfile = open(outfilename, mode)
        self.stdout_also = stdout_also
        self.line_count = 0

    def get_count(self):
        return self.line_count

    def write(self, content):
        self.outfile.write(content)
        if self.stdout_also:
            sys.stdout.write(content)
        count = content.count("\n")
        self.line_count += count

    def close(self):
        self.outfile.close()


# used to sort register name according to their block
def register_keys(text):
    if text[0] == "9":  # block 9 registers come at last
        return ["Z", text[1:3]]
    else:
        return [atoi(c) for c in re.split(r"(\d+)", text)]


def compare_register_fields(module_name, registers):
    "compares register_fields found scanning the PDF with python-sped when possible"
    registers_lib = Escrituracao(module_name, 2017)._registros
    error_count = 0
    field_count = 0
    for register_info in registers:
        register_name = register_info["code"]
        register_lib = getattr(registers_lib, "Registro{}".format(register_name))
        lib_fields = [f.nome for f in register_lib.campos if f.nome != "REG"]
        # TODO move field extraction outside from method?
        headers, rows = extract_fields_spec(module_name, register_name)
        register_info["spec_fields"] = rows
        spec_fields = [f["code"] for f in rows]
        extra_spec = list(set(spec_fields) - set(lib_fields))
        extra_lib = list(set(lib_fields) - set(spec_fields))
        if len(extra_spec) > 0 or len(extra_lib) > 0:
            print("\n{}".format(register_name))
            if len(extra_lib) > 0:
                print("     missing in pdf: %s" % (extra_lib))
            if len(extra_spec) > 0:
                print("     missing in python-sped: %s" % (extra_spec))
            error_count += len(extra_spec) + len(extra_lib)
        field_count += len(lib_fields)
        # TODO also report field types error (very likely inside python-sped)
    print(
        "\n %s field errors out of %s\
          fields (%s percent)"
        % (error_count, field_count, 100.0 / (field_count / float(error_count)))
    )


def collect_register_children(registers):
    "reads the registers hierarchy"
    for register_info in registers:
        if register_info.get("spec") and register_info["spec"]["level"] > 1:
            collect_children = False
            children_o2m = []
            children_m2o = []
            level = register_info["spec"]["level"]
            for r in registers:
                if r == register_info:
                    collect_children = True
                    continue
                if collect_children:
                    if r.get("spec") and r["spec"]["level"] == level + 1:
                        if "1:1" in r["spec"]["card"] or "1;1" in r["spec"]["card"]:
                            children_m2o.append(r)
                        else:
                            children_o2m.append(r)
                            child_specs = list(
                                filter(
                                    lambda x: x["code"] == r["spec"]["code"], registers
                                )
                            )
                            if child_specs:
                                child_specs[0]["parent"] = register_info
                    elif r.get("spec") and r["spec"]["level"] <= level:
                        break
            register_info["children_o2m"] = children_o2m
            register_info["children_m2o"] = children_m2o
            if PRINT_HIERARCHY:
                if children_o2m:
                    print("  " * level + "children o2m:")
                    for c in children_o2m:
                        print("  " * level + "  " + c["code"])
                if children_m2o:
                    print("  " * level + "children m2o:")
                    for c in children_m2o:
                        print("  " * level + "  " + c["code"])


def generate_model(options, module_name):
    "entry point to generate all models and views"
    print("\n\n***** GENERATING CODE FOR ", module_name)
    models_file_name = "../models/{}.py".format(module_name)
    views_file_name = "../views/{}.xml".format(module_name)

    if (os.path.exists(models_file_name)) and not options.force:
        sys.stderr.write("\nfiles exist already.  " "use -f/--force to overwrite.\n\n")
        sys.exit(1)
    models_writer = Writer(models_file_name)
    views_writer = Writer(views_file_name)
    security_writer = Writer("../security/{}".format("ir.model.access.csv"), mode="a")

    wrtmodels = models_writer.write
    wrtviews = views_writer.write
    wrtsecurity = security_writer.write

    wrtmodels("# -*- coding: utf-8 -*-\n\n")
    wrtmodels("from odoo import models, fields\n\n")
    wrtmodels("from . import spec_models\n\n")

    wrtviews('<?xml version="1.0" encoding="UTF-8"?>\n<odoo>')
    wrtviews('\n    <menuitem name="SPED" id="menu_root" sequence="115"/>')
    wrtviews(
        ('\n    <menuitem name="%s"' ' parent="menu_root" id="%s" sequence="2" />')
        % (module_name.upper(), module_name.lower())
    )

    last_bloco = None
    registers = []
    # collect registers
    registers_spec = extract_registers_spec(module_name)

    r_spec_l = {r["code"] for r in registers_spec}
    actual_registers = r_spec_l

    registers_lib = None
    try:
        registers_lib = Escrituracao(module_name, 2017)._registros
        sorted_registers_lib = sorted(dir(registers_lib), key=natural_keys)
        r_lib_l = {
            name[8:12]
            for name in sorted_registers_lib
            if "Registro" in name and name != "Registro"
        }
        actual_registers = r_lib_l
        print("registers not found in pdf: %s" % (list(r_lib_l - r_spec_l,)))
        print("registers not found in python-sped: %s" % (list(r_spec_l - r_lib_l,)))

    except:
        print("\nWARNING python-sped lib for {} cannot be loaded!".format(module_name))
        print("so there will be no checking against the python-sped layout!\n")

    for register_name in sorted(actual_registers, key=register_keys):
        # skip opening and ending registers:
        if register_name[1:4] == "001" or register_name[1:4] == "990":
            continue
        register_specs = list(
            filter(lambda r: r["code"] == register_name, registers_spec)
        )
        register_spec = register_specs and register_specs[0] or None
        headers, rows = extract_fields_spec(module_name, register_name)

        registers.append(
            {
                "code": register_name,
                "spec": register_spec,
                "spec_fields": rows,
                "children_o2m": [],
                "children_m2o": [],
            }
        )

    if registers_lib != None:
        compare_register_fields(module_name, registers)
    collect_register_children(registers)

    # generate models
    for register_info in registers:
        spec_fields = register_info["spec_fields"]
        if len(spec_fields) < 1:
            # TODO check errors. Then skip if no field?
            print(
                "NO FIELD FOUND IN REGISTER {}, SKIPPING IT".format(
                    register_info["code"]
                )
            )
        register_name = register_info["code"]
        register_specs = list(
            filter(lambda r: r["code"] == register_name, registers_spec)
        )
        if len(register_specs) == 0:
            print("SKIPPING REG {} AS NO (PDF) SPEC FOUND".format(register_name))
            continue
        register_spec = register_specs and register_specs[0] or None
        wrtmodels("\nclass Registro{}(models.Model):\n".format(register_name))
        wrtmodels(
            "    _name = 'l10n.br.sped.%s.%s'\n"
            % (module_name.lower(), register_name.lower(),)
        )
        wrtmodels(
            '    _description = u"""{}"""\n'.format(
                register_spec["desc"] or register_name
            )
        )
        wrtmodels("    _inherit = 'l10n.br.sped.mixin'\n")

        wrtsecurity(
            "access_%s_%s,%s.%s,model_%s_%s_%s,base.group_user,1,1,1,1\n"
            % (
                module_name.lower(),
                register_name.lower(),
                module_name.lower(),
                register_name.lower(),
                "l10n_br_sped",
                module_name.lower(),
                register_name.lower().replace(".", "_"),
            )
        )

        bloco_char = register_name[0]
        if bloco_char != last_bloco:
            # TODO put back bloco desc in menuitem name after extracting it
            wrtviews(
                ('\n\n    <menuitem name="Bloco %s"' ' parent="%s" id="%s_%s"/>')
                % (bloco_char, module_name.lower(), module_name, bloco_char.lower())
            )
        last_bloco = bloco_char

        #        for field in obj.campos:
        for spec in spec_fields:
            name, label, help_msg = get_name_label_help(spec)

            if spec["type"] == "D":
                if spec.get("required", False):
                    wrtmodels(
                        '    {} = fields.Date("{}", required=True)\n'.format(
                            name, label
                        )
                    )
                else:
                    wrtmodels('    {} = fields.Date("{}")\n'.format(name, label))
            elif spec.get("type") == "N" or name.startswith(
                "vl_"
            ):  # TODO use spec_type info too
                if spec.get("required", False):
                    if spec.get("digits"):
                        wrtmodels(
                            '    %s = fields.Monetary("%s", required=True, digits=%s)\n'
                            % (name, label, spec.get("digits"))
                        )
                    else:  # if spec.get('int_size'):
                        wrtmodels(
                            '    %s = fields.Integer("%s", required=True)\n'
                            % (name, label)
                        )  # TODO or Char with Select sometimes?
                else:
                    if spec.get("digits"):
                        wrtmodels(
                            '    %s = fields.Monetary("%s", digits=%s)\n'
                            % (name, label, spec.get("digits"))
                        )
                    else:  # spec.get('int_size'):
                        wrtmodels(
                            '    {} = fields.Integer("{}")\n'.format(name, label)
                        )  # TODO or Char with Select sometimes?
            else:
                if name == "":
                    name = "reg_av"
                    label = "REG_AV"  # FIXME ECF Registro9100 undefined field; brittle
                if spec.get("required", False):
                    wrtmodels(
                        '    %s = fields.Char("%s", '
                        "required=True,"
                        '\n        help="""%s""")\n' % (name, label, help_msg)
                    )
                else:
                    wrtmodels(
                        '    %s = fields.Char("%s",'
                        '\n        help="""%s""")\n' % (name, label, help_msg)
                    )
            # TODO Bolean fields, if values are [S;N], see python-sped
        if register_info.get("parent"):
            name = register_info["parent"]["spec"]["code"].lower()
            label = register_info["parent"]["spec"]["desc"]
            label = label.strip().splitlines()[0]
            wrtmodels(
                "    parent_%s_id = fields.Many2one('l10n.br.sped.%s.%s',\n\
                                     string=\"%s\")\n"
                % (name, module_name.lower(), name, label)
            )

        for child in register_info["children_m2o"]:
            name = child["spec"]["code"].lower()
            # TODO if label is not short move to help
            label = child["spec"]["desc"].strip()
            label = label.strip().splitlines()[0]
            help = "Bloco {}".format(child["spec"]["code"][0])
            wrtmodels(
                "    reg_%s_id = fields.Many2one('l10n.br.sped.%s.%s',\n\
                              string=\"%s\",\n\
                              help='%s') # m2o\n"
                % (name, module_name.lower(), name, label, help)
            )
        for child in register_info["children_o2m"]:
            name = child["spec"]["code"].lower()
            # TODO if label is not short move to help
            label = (child["spec"]["desc"]).strip()
            label = label.strip().splitlines()[0]
            help = "Bloco {}".format(child["spec"]["code"][0])
            wrtmodels(
                "    reg_%s_ids = fields.One2many('l10n.br.sped.%s.%s',"
                "'parent_%s_id',\n\
                               string=\"%s\",\n\
                               help='%s')\n"
                % (name, module_name.lower(), name, register_name.lower(), label, help)
            )

        if register_info["spec"].get("level") in [0, 1, 2]:  # , "3"]:
            action_name = strip_desc(register_info["spec"]["desc"])
            action = """\n
    <record id='%s_%s_action' model='ir.actions.act_window'>
        <field name="name">%s</field>
        <field name="res_model">l10n.br.sped.%s.%s</field>
        <field name="view_type">form</field>
        <field name="view_mode">tree,form</field>
    </record>""" % (
                module_name,
                register_name.lower(),
                action_name,
                module_name.lower(),
                register_name.lower(),
            )
            wrtviews(action)
            wrtviews(
                ('\n    <menuitem action="%s_%s_action"' ' parent="%s_%s" id="%s_%s"/>')
                % (
                    module_name,
                    register_name.lower(),
                    module_name,
                    bloco_char.lower(),
                    module_name,
                    register_name.lower(),
                )
            )

    wrtviews("\n</odoo>")
    models_writer.close()
    views_writer.close()


def get_name_label_help(field):
    name = unidecode(field["code"]).encode("ascii").decode("utf8")

    name = name.replace(" ", "_").replace("-", "_").replace(";", "").replace("*", "")
    if len(field["desc"]) < 40:
        label = field["desc"]
    else:
        label = field["desc"].split(".")[0]
        if len(label.split(":")[0]) > 15:
            label = label.split(":")[0]
        if len(label) > 64:
            label = name
    label = label.replace('"', "'")
    if field.get("conditional_required"):
        label += "*"
    # TODO help. Also put the pdf page in the help
    help_msg = "{}\nVer pagina {}".format(field["desc"], field["page"])
    #    help_msg = help_msg.replace('"', "'")
    return name.lower(), label, help_msg


def strip_desc(desc):
    desc = desc.strip().splitlines()[0]
    if len(desc) > 64:
        s = desc.split("(")
        if len(s) > 1 and len(s[0]) < 64:
            return s[0]
        s = desc.split("-")
        if len(s) > 30 and len(s[0]) < 64:
            return s[0]
        s = desc.split(",")
        if len(s) > 30 and len(s[0]) < 64:
            return "{},...".format(s[0])
        s = desc.split(" ")
        desc2 = ""
        for w in s:
            if len(desc2) + len(w) < 61:
                desc2 = "{} {}".format(desc2, w)
            else:
                break
        return "{}...".format(desc2)
    else:
        return desc


USAGE_TEXT = __doc__


def usage():
    print(USAGE_TEXT)
    sys.exit(1)


def main():
    args = sys.argv[1:]
    try:
        opts, args = getopt.getopt(
            args, "hfs:", ["help", "force", "no-class-suffixes",]
        )
    except:
        usage()
    options = ProgramOptions()
    options.force = False
    options.class_suffixes = False
    for opt, val in opts:
        if opt in ("-h", "--help"):
            usage()
        elif opt in ("-f", "--force"):
            options.force = True
        elif opt == "--no-class-suffixes":
            options.class_suffixes = False
    generate_model(options, "ecd")
    generate_model(options, "ecf")
    generate_model(options, "efd_icms_ipi")
    generate_model(options, "efd_pis_cofins")
    # generate_model(options, 'fci')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
This module is divided in 5 steps in order to build usable CSV files from the raw CSV
extracted by extract_tables.py :

1.  `extract_registers_list` extracts a list of the module's registers
    from the module's blocks registers lists.
2.  `extract_register_fields` extracts fields rows from one register's table as they
    appear in the raw CSV files.
3.  Using this method and the module's registers list, `build_accurate_fields_csv`
    creates a CSV file named 'MODULE_accurate_fields.csv' listing the fields rows
    for each module as they appear in the original pdf.
4.  From these 'accurate CSV files', two methods `get_fields` and `get_registers` are
    defined and return lists of the module's fields and registers made as usable python
    dictionaries with 'interpreted' information (field's type, booleans for required
    fields, register's level...)
5. `build_registers_csv` and `build_usable_fields_csv` finally create CSV files listing
    the module's registers and fields from these 'interpreted' dictionaries.

"""

import csv
import logging
import re

import click

from . import extract_tables
from .constants import MODULE_HEADER, MODULES, MOST_RECENT_YEAR, OLDEST_YEAR, SPECS_PATH

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def _get_mod_header(mod):
    # Override this method if the hard code MODULE_HEADER is not wanted
    return MODULE_HEADER.get(mod)


# used to sort csv files
def _atoi(text):
    return int(text) if text.isdigit() else text


# used to sort csv files
def natural_keys(file):
    """
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    """
    return [_atoi(c) for c in re.split(r"(\d+)", file.name)]


def get_raw_rows(mod, year):
    """Walk through ./specs/YEAR/MODULE/raw_camelot_csv/ and return a big dictionary
    of all the raw rows found in raw CSV files extracted by ./extract_tables.py,
    gathered by their page number :

    raw_rows = {
        12: [
            ["Empresas  Obrigad...", "Devem  produzir  o  livro  “Z”..."],
            ["Nº", "Campo", "Descrição", "Tipo", "Tamanho", "Decimal"],
            ["01", "REG", "Identificador do registro: I510", "C", "004", "-"],
            ["02", "NAT_SUB_CNT", "Natureza  da  subconta", "C", "002", "-"],
            ["03", "COD_SUB_CNT", "Código da subco....", "C", "020", "-"],
            ...
        ],
        13: [[...], [...]],
        ...
    }
    """
    raw_rows = {}
    path_raw = SPECS_PATH / f"{year}" / f"{mod}" / "raw_camelot_csv"

    if not path_raw.exists():
        extract_tables.extract_mod_tables(mod, year)

    csv_files = sorted([file for file in path_raw.iterdir()], key=natural_keys)

    for csv_file in csv_files:
        page = int(csv_file.name.split("-")[2])
        with open(csv_file, "r") as csvfile:
            reader = csv.reader(csvfile, delimiter=",", quotechar='"')
            for row in reader:
                raw_rows.setdefault(page, []).append(row)

    assert path_raw, (
        f"No raw CSV files found at '{str(path_raw.resolve())}/'.\n"
        "Run 'python -m spedextractor.extract_tables' before continuing."
    )

    return raw_rows


def clean_row(row):
    """Clean row's content"""
    row = [str(x) for x in row]
    for index, cell in enumerate(row):
        # replace all the whitespaces ("  ", \n, \r...)
        clean_cell = " ".join(cell.split())
        # e.g. change "Entr." to "Entr" in fields table's headers
        if re.match(r"^[a-zA-Z]+\.$", clean_cell):
            clean_cell = clean_cell[:-1]
        # e.g. change "N’" to "N" in column "Tipo"
        if re.match(r"^[a-zA-Z]+\’$", clean_cell):
            clean_cell = clean_cell[:-1]
        row[index] = clean_cell

    return row


# 1. `extract_registers_list`
# ===========================


def _is_register_code(code):
    return code and len(code) == 4 and code[1:3].isdigit()


def _map_register_row(mod, row):
    """Extracts register's row information for each kind of module"""
    # TODO : Join rows content when they are from the same register's line but
    # splited in two because of page break.
    # Example : EFD ICMS IPI pdf Outubro 2019 p20-21
    register = {}
    if len(row) > 2:
        if mod == "ecd":
            if len(row[0]) == 1 and row[0] != "":
                register = {
                    "block": row[0],
                    "code": row[2],
                    "desc": row[1],
                    "level": row[3],
                    "card": row[4],
                }
        elif mod == "ecf":
            if len(row[0]) == 4:
                register = {
                    "block": row[0][0],
                    "code": row[0],
                    "desc": row[2],
                    "level": row[1],
                    "card": row[5],
                }
        elif mod == "efd_pis_cofins":
            code = row[2][:-4] if row[2].endswith(" (*)") else row[2]
            if len(code) == 4 and len(row) >= 5:
                register = {
                    "block": row[0],
                    "code": code,
                    "desc": row[1],
                    "level": row[3],
                    "card": row[4],
                }

        elif mod == "efd_icms_ipi":
            if len(row[0]) == 1 and row[0] != "":
                register = {
                    "block": row[0],
                    "code": row[2],
                    "desc": row[1],
                    "level": row[3],
                    "card": row[4],
                }

    if register.get("code") and _is_register_code(register["code"]):
        if str(register["level"]).isdigit():
            register["level"] = int(register["level"])
        else:
            return False
    else:
        return False

    return register


def extract_registers_list(mod, year, raw_rows=None):
    """Scans the raw csv rows and return 'registers', a list of dictionaries giving
    all the information about the module's registers (block, code, description,
    hierarchy level and card) found in the block's registers lists."""
    registers = []
    in_block = False
    if not raw_rows:
        raw_rows = get_raw_rows(mod, year)

    for page in raw_rows:
        for row in raw_rows[page]:
            if (
                ("BLOCO" in row or "Bloco" in row or "Registro" in row)
                and (
                    "NÍVEL" in row
                    or "Nível" in row
                    or r"N\xc3\xadvel" in row
                    or "Nome do Registro" in row
                    or "Reg." in row
                )
                or "BLOCO  DESCRIÇÃO" in row  # ecf
            ):  # ecd
                in_block = True
                continue
            if in_block:
                if "".join(row) == "":
                    continue  # empty line
                reg = _map_register_row(mod, clean_row(row))
                if reg and reg["code"] not in [r["code"] for r in registers]:
                    registers.append(reg)

                    if reg["code"] == "9999":
                        return registers

    return registers


# 2. `extract_register_fields`
# ============================


def _is_joined_index(row, c):
    """ Checks if row's column 'c' start with row's index and need to be split"""
    if len(row) > 4 and row[c][0:2].isdigit() and len(row[c]) > 5 and " " in row[c]:
        return True
    return False


def _split_code_desc(row, c):
    """Checks if row's column 'c' is a joined code and description and return 2 split
    items to be used if true"""
    i_end = 0
    code = []
    desc = []
    for i, part_cell in enumerate(row[c].split(" ")):
        if not re.match(r"\b[A-Z_ÇÃÕÍÚe0-9]+\b", part_cell):
            i_end = i
            break
    if i_end != 0:
        split = row[c].split(" ")
        code = " ".join(split[:i_end])
        desc = " ".join(split[i_end:])
        return code, desc
    else:
        return None


def _format_row(row):
    """Separates columns joined together"""
    # change ["04  VL_BC_RET", ""] into ["04","VL_BC_RET"]
    if _is_joined_index(row, 0) and row[1] == "":
        split = row[0].split(" ")
        row = [split[0], " ".join(split[1:])] + row[2:]
    # change ["", "04  VL_BC_RET"] into ["04","VL_BC_RET"]
    if _is_joined_index(row, 1) and row[0] == "":
        split = row[1].split(" ")
        row = [split[0], " ".join(split[1:])] + row[2:]

    # change ["02","NUM_ITEM  Número seqüencial do item no documento fiscal","", "N"]
    # into ["02","NUM_ITEM", "Número seqüencial do item no documento fiscal", "N"]
    if _split_code_desc(row, 1) and row[2] == "":
        row[1], row[2] = _split_code_desc(row, 1)

    # fix ["02","", "NUM_ITEM  Número seqüencial do item no documento fiscal", "N"]
    if _split_code_desc(row, 2) and row[1] == "":
        row[1], row[2] = _split_code_desc(row, 2)

    # fix ["02", "NUM_ ITEM  Número seqüencial do item no documento fiscal", "N"]
    if _split_code_desc(row, 1) and row[2] != "":
        row[1], desc = _split_code_desc(row, 1)
        row.insert(2, desc)

    # Sommetimes the description cell (=row[2]) itself is split and override type's cell
    # (=row[3])
    if len(row[3]) > 2:
        row[2] = row[2] + " " + row.pop(3)

    return row


def _map_row_mod_header(row, mod):
    """Inserts empty column when needed to align with module's header columns order"""
    len_header = len(_get_mod_header(mod))
    if row and mod == "efd_icms_ipi":
        if len(row) == len_header - 1:
            # i.e. row has the columns 'Entr' and 'Saída' but not 'Obrig'
            row.insert(6, "")
    # Add empty cells in row if incomplete
    if row and len(row) < len_header:
        extension = [""] * (len_header - len(row))
        row.extend(extension)

    return row


def _is_reg_row(row):
    if "REG" in row[1] and "Texto" in row[2]:
        return True
    return False


def _apply_camelot_patch(mod, year, register, row):
    """Catches patched row in ./camelot_patch/ and return override current row"""
    patch_file = SPECS_PATH / f"{year}" / "camelot_patch" / f"{mod}_camelot_patch.csv"

    try:
        with open(patch_file, "r") as patch_csv:
            patch_rows = csv.reader(patch_csv, delimiter=",", quotechar='"')
            for patch_row in patch_rows:
                # N-B : A patch can be applied twice if a row which is not a register's
                # field match this condition (e.g. field TIP_ENT in ECF register 0010).
                # But it is not a problem as the non-field row will be ignore after.
                if patch_row[0] == register and patch_row[2] == row[0]:
                    row = patch_row[2:]
                    logger.info(f"    PATCH : field {row[1]} in register {register}")
    except FileNotFoundError:
        return row
    return row


def _is_field_row(row, last_field_index):
    """Returns True if the row match a series of condition to be a register's field"""
    if (
        row[1] != ""
        and len(row[1]) < 32
        and len(row[1]) > 1
        and not row[1][0].isdigit()
        and "RZ_CONT" not in row[1]  # in ECD, doesn't look like a real data field
        and (
            row[0].isdigit()
            and row[1] != "REG"
            and int(row[0]) == last_field_index + 1
            or row[0] == "*"
        )
    ):
        return True
    else:
        return False


def extract_register_fields(mod, year, register_name, raw_rows=None, patch=True):
    """Scans the raw_rows to find the rows describing the fields
    of a given register."""
    in_register = False
    last_field_index = 1
    reg_fields = []

    if not raw_rows:
        raw_rows = get_raw_rows(mod, year)

    for page in raw_rows:
        for row in raw_rows[page]:
            if "".join([str(x) for x in row]) == "" or len(row) < 4:
                continue  # empty line

            row = _format_row(clean_row(row))
            row = _map_row_mod_header(row, mod)
            # N-B : We assume losing performance by cleaning and aligning every row
            # instead of only the worthy ones in order to simplify the identification
            # of register's fields rows (realized by _is_reg_row and _is_field_row).

            if not in_register and _is_reg_row(row) and register_name in row[2]:
                # We found the register's table first row describing the register
                # itself
                in_register = True
                # Add register's name and page columns
                row.insert(0, page)
                row.insert(0, register_name)
                reg_fields.append(row)
                continue

            if in_register:
                if _is_reg_row(row) and register_name not in row[2]:
                    # next register table found -> stopping
                    return reg_fields

                if patch:
                    row = _apply_camelot_patch(mod, year, register_name, row)

                # TODO : handle instances where the field's row is split in two by a
                # page break. (=all the fields are empty except Description - 3rd
                # column). Example : EFD PIS COFINS page 78 Registro 0200
                if _is_field_row(row, last_field_index):
                    last_field_index = int(row[0])

                    # Add register's name and page columns
                    row.insert(0, page)
                    row.insert(0, register_name)
                    reg_fields.append(row)

    return reg_fields


# 3. `build_accurate_fields_csv`
# ==============================


def build_accurate_fields_csv(
    mod, year, raw_rows=None, extracted_registers=None, patch=True
):
    """Build a CSV file with the module's fields rows as they appear in the original
    pdf.

    If the registers list is passed as an argument, it avoids to make the extraction
    another time."""
    accurate_file = SPECS_PATH / f"{year}" / f"{mod}" / f"{mod}_accurate_fields.csv"
    reg_with_no_field = []
    if not extracted_registers:
        extracted_registers = extract_registers_list(mod, year, raw_rows)

    logger.info(f"> Building {mod}_accurate_fields.csv")

    # Catch all the module's fields
    mod_fields = []
    for register in extracted_registers:
        reg_fields = extract_register_fields(
            mod, year, register["code"], raw_rows, patch
        )
        mod_fields.extend(reg_fields)
        # Catch registers with no fields for information.
        # N-B : A register table must have at least two rows (the first row describes
        # the register itself)
        if len(reg_fields) < 2:
            reg_with_no_field.append(register["code"])

    if reg_with_no_field:
        logger.info(
            f"    /!\\ {len(reg_with_no_field)} registers with no field "
            f"catched by camelot : {reg_with_no_field}"
        )

    with open(accurate_file, "w") as accurate_csv:
        # Delete actual fields_file's datas before writing
        accurate_csv.seek(0)
        accurate_csv.truncate()

        accurate_rows = csv.writer(
            accurate_csv, delimiter=",", quotechar='"', quoting=csv.QUOTE_ALL
        )
        # Write module's header
        mod_header = ["Register", "Page"] + [c[0] for c in _get_mod_header(mod)]
        accurate_rows.writerow(mod_header)
        # Write module's fields
        for row in mod_fields:
            accurate_rows.writerow(row)


# 4. `get_fields` and `get_registers`
# ===================================


def _normalize_field_code(code):
    return code.replace("  ", "").replace(" ", "").replace("__", "_")


def _convert_field_type(field):
    """Return a string giving the 'interpreted' field's type :
    'char', 'int', 'float' or 'date'. """
    spec_type = field.get("spec_type")
    if spec_type:
        code = field["code"]
        if spec_type == "D" or code.startswith("DT_") or code.startswith("DATA_"):
            field["type"] = "date"
        elif spec_type == "N":
            field["type"] = "float" if field.get("decimal") else "int"
        # If no given type, define it as "character"
        # "NS" ("Numérico Com Sinal") means that the field's value must be "+" ou "-"
        # cf. ECF pdf page 26
        elif not spec_type or spec_type in ["C", "NS"]:
            field["type"] = "char"
        else:
            logger.warning(
                f"    Could not define field {field['code']} type in register "
                f"{field['register']}"
            )
    return field


def _convert_field_required(field):
    """Return field with additional required boolean keys if necessary"""
    spec_required = field.get("spec_required")
    if spec_required in ["O", "S", "Sim", "Sm", "sim"]:
        field["required"] = True
    elif spec_required in ["N", "Não"]:
        field["required"] = False
    elif spec_required == "OC":
        field["conditional_required"] = True
    elif field.get("register") and field.get("spec_required"):
        logger.warning(
            f"    Could not define if field {field['code']} is required in register "
            f"{field['register']}"
        )
    elif field.get("spec_required"):
        # In this case we are converting a register item not a field
        logger.warning(f"    Could not define if register {field['code']} is required")
    return field


def _convert_field_in_out(field):
    # TODO : interpret field["spec_in"] when it is an integer
    # (cf register C170 in EFD_ICMS_IPI page 71)
    spec_in = field.get("spec_in")
    if spec_in == "O":
        field["in_required"] = True
    elif spec_in == "OC":
        field["conditional_in_required"] = True

    spec_out = field.get("spec_out")
    if spec_out == "O":
        field["out_required"] = True
    elif spec_out == "OC":
        field["conditional_out_required"] = True
    return field


def _convert_values(field):
    """Add a 'values' keys if field["spec_values"] can be interpreted as a list of items
    """
    values = field.get("spec_values")
    if values:
        # Remove unnecessary quotes
        values = field["spec_values"].replace("”", "").replace("“", "").replace('"', "")
        if values[0] == "[" and values[-1] == "]":
            values = values[1:-1]
        if "," in values:
            values = values.split(",")
        elif ";" in values:
            values = values.split(";")
        if type(values) == list:
            field["values"] = [v.replace(" ", "").replace("''", "") for v in values]
        if field.get("spec_type") and field["spec_type"] == "NS":
            field["values"] = ["+", "-"]  # cf. ECF pdf page 26
    # TODO : There is still around 30 fields with "spec_values" which is not easily
    # convertible into a list
    return field


def _convert_rules(field):
    """Convert rules string in a iterable list"""
    rules = field.get("rules")
    if rules:
        field["rules"] = rules[:-1].replace(" ", "").replace("[", "").split("]")
    return field


def _map_field_row(row, mod):
    """Return a field dictionary with interpreted information"""
    field = {}
    mod_keys = [c[1] for c in _get_mod_header(mod)]

    # Catch raw datas from row
    field["register"] = row[0]
    for index, key in enumerate(mod_keys):
        if index + 2 < len(row) and row[index + 2] not in ["-", ""]:
            field[key] = row[index + 2]

    # Interpret raw datas
    field["index"] = int(field["index"].replace("*", "0"))  # TODO check * cases
    field["code"] = _normalize_field_code(field["code"])

    field = _convert_field_type(field)
    field = _convert_field_required(field)
    field = _convert_field_in_out(field)
    field = _convert_values(field)
    field = _convert_rules(field)

    return field


def get_fields(mod, year=MOST_RECENT_YEAR, with_reg=False):
    """Returns a list of the module's fields recorded as dictionaries with interpreted
    values.

    :param mod: The module from which are catched the fields.
    :type mod: str

    :param year: The module's pdf year version.
    :type year: int

    :param with_reg: An optional argument to add the REG field (opening every registers
    tables) to this list of dictionaries. Used in build_pythonsped_json.py.
    :type with_reg: bool
    """

    if mod not in MODULES:
        raise ValueError(
            f"'{mod}' is not a valid module name. Choose between {MODULES}"
        )
    accurate_file = SPECS_PATH / f"{year}" / f"{mod}" / f"{mod}_accurate_fields.csv"
    fields = []

    # Build MODULE_accurate_fields.csv if empty or not existing
    if not accurate_file.exists() or accurate_file.stat().st_size == 0:
        build_accurate_fields_csv(mod, year)

    # Open the CSV with the accurate fields list
    with open(accurate_file, "r") as accurate_csv:
        accurate_rows = csv.reader(accurate_csv, delimiter=",", quotechar='"')
        # Avoid CSV header
        next(accurate_rows)
        for row in accurate_rows:
            if row[3] != "REG" or with_reg:
                fields.append(_map_field_row(row, mod))

    logger.info(f"    {len(fields)} fields catched in {mod.upper()}")
    return fields


def get_registers(mod, year=MOST_RECENT_YEAR, raw_rows=None, extracted_registers=None):
    """Add the `required` and `field_in_out` attributes (calculated from the
    MODULE_accurate_fields.csv file) to the registers extracted by
    `extract_registers_list()` and return this registers list.

    If the `extracted_registers` is passed as an argument, it avoids to make the
    extraction another time."""
    if mod not in MODULES:
        raise ValueError(
            f"'{mod}' is not a valid module name. Choose between {MODULES}"
        )
    accurate_file = SPECS_PATH / f"{year}" / f"{mod}" / f"{mod}_accurate_fields.csv"
    mod_keys = [c[1] for c in _get_mod_header(mod)]

    if not raw_rows:
        raw_rows = get_raw_rows(mod, year)
    if extracted_registers:
        registers = extracted_registers
    else:
        registers = extract_registers_list(mod, year, raw_rows)

    # Build MODULE_accurate_fields.csv if empty or not existing
    if not accurate_file.exists() or accurate_file.stat().st_size == 0:
        build_accurate_fields_csv(mod, year, raw_rows, registers)

    # Open the CSV with the accurate fields list
    with open(accurate_file, "r") as accurate_csv:
        accurate_rows = csv.reader(accurate_csv, delimiter=",", quotechar='"')
        # Avoid CSV header
        next(accurate_rows)
        for row in accurate_rows:
            if row[3] == "REG":
                for register in registers:
                    if register["code"] == row[0]:
                        # Catch row's 'spec_required', 'spec_in' and 'spec_out'
                        for spec in ["spec_required", "spec_in", "spec_out"]:
                            if spec in mod_keys:
                                index = mod_keys.index(spec) + 2
                                if index < len(row) and row[index] not in ["-", ""]:
                                    register[spec] = row[index]

                        register = _convert_field_required(register)
                        register = _convert_field_in_out(register)
                        break

    logger.info(f"    {len(registers)} registers catched in {mod.upper()}")
    return registers


# 5. `build_usable_fields_csv` and `build_registers_csv`
# ======================================================


def _sort_header_order(key):
    """Reorder hearder's keys"""
    header_base = [
        "register",
        "index",
        "block",
        "code",
        "type",
        "required",
        "in_required",
        "out_required",
        "values",
        "rules",
    ]
    if key == "desc":
        return 99
    elif key.startswith("spec_"):
        return 90
    elif key.startswith("conditional_"):
        return 10
    elif key in header_base:
        return header_base.index(key)
    else:
        return 50


def _get_usable_csv_header(fields):
    """Return a list of all the different keys available in fields"""
    header = []
    for field in fields:
        for key in field.keys():
            if key not in header:
                header.append(key)
    header.sort(key=_sort_header_order)
    return header


def build_usable_fields_csv(mod, year):
    fields_file = SPECS_PATH / f"{year}" / f"{mod}" / f"{mod}_fields.csv"
    fields = get_fields(mod, year)

    logger.info(f"> Building {mod}_fields.csv")

    # Open the CSV with the accurate fields list
    with open(fields_file, "w") as f_file:
        # Delete actual usable_file's datas before writing
        f_file.seek(0)
        f_file.truncate()

        fields_csv = csv.writer(
            f_file, delimiter=",", quotechar='"', quoting=csv.QUOTE_ALL
        )

        header = _get_usable_csv_header(fields)
        fields_csv.writerow(header)
        for field in fields:
            row = []
            # Add missing keys with empty values to field
            for col in header:
                if col not in field.keys():
                    field[col] = ""
                row.append(field[col])
            fields_csv.writerow(row)


def build_registers_csv(mod, year, raw_rows=None, extracted_registers=None):
    """Generate a csv with the Registers specifications. One line for each register.
    If no registers argument is passed, the registers list extraction will be made by
    `get_registers`.
    """
    registers_file = SPECS_PATH / f"{year}" / f"{mod}" / f"{mod}_registers.csv"
    registers = get_registers(mod, year, raw_rows, extracted_registers)
    header = _get_usable_csv_header(registers)

    logger.info(f"> Building {mod}_registers.csv")

    with open(registers_file, "w") as reg_file:
        # Delete actual reg_file's datas before writing
        reg_file.seek(0)
        reg_file.truncate()

        reg_csv = csv.writer(
            reg_file, delimiter=",", quotechar='"', quoting=csv.QUOTE_ALL
        )
        # First line is columns titles
        reg_csv.writerow(header)
        for register in registers:
            row = []
            # Add missing keys with empty values to field
            for col in header:
                if col not in register.keys():
                    register[col] = ""
                row.append(register[col])
            reg_csv.writerow(row)


# 6. BONUS : `get_blocks` to get module's blocks list
# ======================================================


def extract_blocks(mod, year, raw_rows=None):
    """Return a list of the module's blocks rows as found in path_raw"""
    extracted_blocks = []
    in_block_list = False

    if not raw_rows:
        raw_rows = get_raw_rows(mod, year)

    for page in raw_rows:
        for row in raw_rows[page]:
            if (
                len(row) in [2, 3]
                and row[0] == "Bloco"
                and row[1] in ["Descrição", "Nome do Bloco"]
            ):
                in_block_list = True
                continue
            if in_block_list and row[0] != "Bloco":
                extracted_blocks.append(row)
                if row[0] == "9":
                    return extracted_blocks
    if not extracted_blocks:
        logger.warning(f"    /!\\ No Blocks list catched in {mod.upper()}")
        return []


def get_blocks(mod, year, raw_rows=None, extracted_blocks=None):
    """Return a list of dictionaries representing module's blocks.
    Used in ./build_python-sped_json.py"""
    if mod not in MODULES:
        raise ValueError(
            f"'{mod}' is not a valid module name. Choose between {MODULES}"
        )
    blocks = []
    if not extracted_blocks:
        extracted_blocks = extract_blocks(mod, year, raw_rows)
    for row in extracted_blocks:
        block = {}
        block["code"] = row[0].replace("*", "")
        block["desc"] = " ".join(row[1].split())
        if len(row) == 3:  # in ECF
            block["info"] = row[2]
        blocks.append(block)
    return blocks


#  main()
# =========================


@click.command()
@click.option(
    "--year",
    default=MOST_RECENT_YEAR,
    show_default=True,
    type=click.IntRange(OLDEST_YEAR, MOST_RECENT_YEAR),
    help="Operate on a specific year's folder, "
    f"can be between {OLDEST_YEAR} and {MOST_RECENT_YEAR}",
)
@click.option(
    "--patch/--no-patch",
    default=True,
    help="Override fields rows extracted by camelot with rows listed in folder "
    "'..specs/YEAR/MODULE/camelot_patch/'",
    show_default=True,
)
def main(year, patch):
    """Build 3 CSV files for each SPED modules (ECD, ECF, EFD_ICMS_IPI and
    EFD_PIS_COFINS) :

    - MODULE_accurate_fields.csv : list all the module's registers fields as they appear
    in the original pdf tables (useful to check the extracted CSV reliability).

    - MODULE_registers.csv : list the module's registers with its useful information.

    - MODULE_fields.csv : list all the module's registers fields with unified and
    'usable' interpreted values (useful to create python objects from these fields)."""

    for mod in MODULES:
        logger.info(f"\n==== Building CSV files for {mod.upper()} {year} ====")

        patch_file = (
            SPECS_PATH / f"{year}" / "camelot_patch" / f"{mod}_camelot_patch.csv"
        )
        if patch and not patch_file.is_file():
            logger.warning(
                f"    No CSV patch '{mod}_camelot_patch.csv' found for {mod.upper()} "
                f"in './specs/{year}/'"
            )

        raw_rows = get_raw_rows(mod, year)
        extracted_registers = extract_registers_list(mod, year, raw_rows)

        build_accurate_fields_csv(mod, year, raw_rows, extracted_registers, patch)
        build_registers_csv(mod, year, raw_rows, extracted_registers)
        build_usable_fields_csv(mod, year)



#!/usr/bin/env python3

import csv
import json
import logging

import click

from .build_csv import get_blocks, get_fields, get_registers
from .constants import MODULES, MOST_RECENT_YEAR, OLDEST_YEAR, SPECS_PATH

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def _convert_length(f):
    """Convert field's length in integer when it is possible"""
    tamanho = f.get("length")
    if tamanho:
        try:
            tamanho = int(tamanho)
        except ValueError:
            pass
    return tamanho


def build_mod_pythonsped_json(mod, year):
    """Build a module dictionary organized with nested registers and fields following
    the structure and the keys for python-sped leiautes and save it as a JSON file"""
    json_file = SPECS_PATH / f"{year}" / f"{mod}" / f"{mod}_pythonsped.json"
    dl_info_file = SPECS_PATH / f"{year}" / "download_info.csv"
    blocks = get_blocks(mod, year)
    registers = get_registers(mod, year)
    fields = get_fields(mod, year, with_reg=True)

    logger.info(f"> Building {mod}_pythonsped.json\n")

    # 1) Initiate dictionary with module's name, date and pdf version taken from the
    # module's download_info.csv file
    module = {}
    module["tipo"] = mod
    with open(dl_info_file, "r") as csvfile:
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')
        header = next(reader)
        col_mod = header.index("module")
        col_version = header.index("version")
        col_date = header.index("date_init")
        for row in reader:
            if row[col_mod] == mod:
                module["versao"] = row[col_version]
                module["data_inicio"] = row[col_date]

    # 2) Add module's blocks list
    module["blocos"] = []
    for b in blocks:
        block = {}
        block["nome"] = b["code"]
        block["descricao"] = b["desc"]
        module["blocos"].append(block)

    # 3) Add module's registers list
    module["registros"] = []
    for r in registers:
        reg = {}
        reg["codigo"] = r["code"]
        reg["nome"] = r["desc"]
        reg["regras"] = []  # Filled after catching 'campos'
        reg["nivel"] = r.get("level")
        reg["ocorrencia"] = r.get("card") if r.get("card") != "1" else "1:1"
        reg["campos_chave"] = []  # TODO Looks like it is part of the fields codes...
        reg["campos"] = []
        for f in fields:
            if f["register"] == r["code"]:
                campo = {}
                campo["indice"] = f["index"]
                campo["nome"] = f["code"]
                campo["descricao"] = f["desc"]
                # Use spec_type not interpreted f["type"]
                campo["tipo"] = f.get("spec_type", "C")
                campo["tamanho"] = _convert_length(f)
                campo["decimal"] = f.get("decimal")
                # Use spec_values not interpreted f["values"]
                campo["valores"] = f.get("spec_values", "-")
                campo["obrigatorio"] = f.get("required", False)
                campo["regras"] = f.get("rules", [])
                # Append campo
                reg["campos"].append(campo)
        # 4) reg['regras']  gathering all the register fields rules
        # TODO it looks like the python-sped leiaute is not following this idea to fill
        # this reg["regras"] item...
        for c in reg["campos"]:
            if c.get("regras"):
                reg["regras"].extend(c.get("regras"))
        # Append reg
        module["registros"].append(reg)

    with open(json_file, "w") as json_f:
        # Delete actual json_f's datas before writing
        json_f.seek(0)
        json_f.truncate()

        json.dump(module, json_f, ensure_ascii=False, indent=2)


@click.command()
@click.option(
    "--year",
    default=MOST_RECENT_YEAR,
    show_default=True,
    type=click.IntRange(OLDEST_YEAR, MOST_RECENT_YEAR),
    help="Operate on a specific year's folder, "
    f"can be between {OLDEST_YEAR} and {MOST_RECENT_YEAR}",
)
def main(year):
    """Build a JSON file with the module's fields for each module."""
    logger.info("Building JSON files for each modules...\n")
    for mod in MODULES:
        build_mod_pythonsped_json(mod, year)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import logging

import click
from sped.efd.icms_ipi import registros as efd_icms_ipi_registers
from sped.efd.pis_cofins import registros as efd_pis_cofins_registers
from sped.escrituracao import Escrituracao

from .build_csv import get_fields, get_registers
from .constants import MODULES, MOST_RECENT_YEAR, OLDEST_YEAR

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def _get_python_sped_reg_and_fields(mod):
    """ Return 2 objects :
    `reg_classes`, a dictionary with python-sped module's registers looking like :
        {
            'F990': <class 'sped.efd.pis_cofins.registros.RegistroF990'>,
            'I001': <class 'sped.efd.pis_cofins.registros.RegistroI001'>,
            'I010': <class 'sped.efd.pis_cofins.registros.RegistroI010'>,
            ...
        }
    and `fields` a list of dictionaries with python-sped module's fields looking like :
        [
            {"register": "F990", "index": 2, "code": "FORMA_APUR_I"},
            {"register": "F990", "index": 3, "code": "IND_ALIQ_CSLL"},
            ...
        ]
    """
    reg_module = {}
    reg_classes = {}
    fields = []

    try:
        # Catch registers defined from 2017 python-sped leiaute
        # (only available for ECD and ECF modules)
        reg_module = Escrituracao(mod, 2017)._registros
    except FileNotFoundError:
        # Catch registers for EFD_ICMS_IPI and EFD_PIS_COFINS modules
        if mod == "efd_icms_ipi":
            reg_module = efd_icms_ipi_registers
        elif mod == "efd_pis_cofins":
            reg_module = efd_pis_cofins_registers

    for name, object in reg_module.__dict__.items():
        if (
            isinstance(object, type)
            and getattr(object, "campos", False)
            and name != "Registro"  # Avoid "Registro" abstract class
        ):
            # Build a reg_classes dict with keys like "M630"
            # taken from registers class names like "RegistroM630"
            reg_classes[name[8:12]] = object

    for reg_name, reg_class in reg_classes.items():
        for field in reg_class.campos:
            if field.nome != "REG":
                fields.append(
                    {"register": reg_name, "index": field.indice, "code": field.nome}
                )

    return reg_classes, fields


def _compare_registers(mod, year, pysped_registers):
    logger.info("\nREGISTERS :")
    ext_registers = get_registers(mod, year)
    ext_reg_codes = [reg["code"] for reg in ext_registers]

    not_in_pysped = [c for c in ext_reg_codes if c not in pysped_registers]
    logger.info(f"    >> {len(not_in_pysped)} not in python-sped : {not_in_pysped}")

    not_in_extractor = [c for c in pysped_registers if c not in ext_reg_codes]
    logger.info(
        f"    >> {len(not_in_extractor)} not in sped-extractor : {not_in_extractor}"
    )

    common_reg = [c for c in ext_reg_codes if c in pysped_registers]
    return common_reg


def _compare_fields(mod, year, common_reg, pysped_fields, detail):
    logger.info("\nFIELDS :")
    ext_fields = get_fields(mod, year)
    not_in_pysped = {}
    not_in_extractor = {}

    for reg in common_reg:
        ext_fields_reg = [f["code"] for f in ext_fields if f["register"] == reg]
        psped_fields_reg = [f["code"] for f in pysped_fields if f["register"] == reg]

        not_in_pysped[reg] = [c for c in ext_fields_reg if c not in psped_fields_reg]
        not_in_extractor[reg] = [c for c in psped_fields_reg if c not in ext_fields_reg]

    logger.info("    Not in python-sped :")
    fields_nb = 0
    reg_nb = 0
    for reg, fields in not_in_pysped.items():
        if fields:
            fields_nb += len(fields)
            reg_nb += 1
            if detail:
                logger.info(f"      {reg} : {fields}")
    logger.info(f"    >> {fields_nb} missing fields in {reg_nb} registers\n")

    logger.info("    Not in sped-extractor :")
    fields_nb = 0
    reg_nb = 0
    for reg, fields in not_in_extractor.items():
        if fields:
            fields_nb += len(fields)
            reg_nb += 1
            if detail:
                logger.info(f"      {reg} : {fields}")
    logger.info(f"    >> {fields_nb} missing fields in {reg_nb} registers\n")


@click.command()
@click.option(
    "--year",
    default=MOST_RECENT_YEAR,
    show_default=True,
    type=click.IntRange(OLDEST_YEAR, MOST_RECENT_YEAR),
    help="Operate on a specific year's folder, "
    f"can be between {OLDEST_YEAR} and {MOST_RECENT_YEAR}",
)
@click.option(
    "--detail/--no-detail",
    default=False,
    show_default=True,
    help="Detail the list of all the fields missing in python-sped or sped-extractor",
)
def main(year, detail):
    """Compare extracted registerts and fields with python-sped library."""
    for mod in MODULES:
        pysped_registers, pysped_fields = _get_python_sped_reg_and_fields(mod)

        logger.info(f"\n-- Comparing {mod.upper()} module...")
        common_reg = _compare_registers(mod, year, pysped_registers)

        _compare_fields(mod, year, common_reg, pysped_fields, detail)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import csv
import logging

import click
import requests

from .constants import MODULES, MOST_RECENT_YEAR, OLDEST_YEAR, SPECS_PATH

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def _get_url(mod, year):
    """Return the first URL found for a module's year version"""
    dl_info_file = SPECS_PATH / f"{year}" / "download_info.csv"
    try:
        with open(dl_info_file, "r") as csvfile:
            reader = csv.reader(csvfile, delimiter=",", quotechar='"')
            header = next(reader)
            col_url = header.index("url")
            col_mod = header.index("module")
            for row in reader:
                if row[col_mod] == mod:
                    return row[col_url]
    except FileNotFoundError:
        logger.exception(
            f"The file '{dl_info_file}' containing the SPED {year} pdfs URLs is "
            "required to download them."
        )


def download_mod_pdf(mod, year):
    logger.info(f"> Downloading pdf {mod.upper()} {year}...")
    pdf_folder = SPECS_PATH / f"{year}" / "pdf"
    # Create pdf folder if necessary
    pdf_folder.mkdir(parents=True, exist_ok=True)

    pdf = pdf_folder / f"{mod}.pdf"
    url = _get_url(mod, year)
    if url:
        try:
            r = requests.get(url)
        except requests.exceptions.MissingSchema:
            logger.exception(
                f"  Cannot download {mod.upper()}'s pdf. '{url}' is not a valid URL."
            )
            pass
        else:
            if not r or r.headers.get("content-type") != "application/pdf":
                logger.warning(f"  No pdf found at '{url}' for {mod.upper()}.")
            else:
                # Remove existing pdf if any
                try:
                    pdf.unlink()
                except FileNotFoundError:
                    pass
                # Create a new one
                with open(pdf, "wb") as f:
                    f.write(r.content)


@click.command()
@click.option(
    "--year",
    default=MOST_RECENT_YEAR,
    show_default=True,
    type=click.IntRange(OLDEST_YEAR, MOST_RECENT_YEAR),
    help="Operate on a specific year's folder, "
    f"can be between {OLDEST_YEAR} and {MOST_RECENT_YEAR}",
)
def main(year):
    """Download SPED specifications pdf from http://sped.rfb.gov.br/"""

    for module in MODULES:
        download_mod_pdf(module, year)


if __name__ == "__main__":
    main()



import pathlib

SPECS_PATH = pathlib.Path(__file__).parent.resolve() / "specs"


def _get_max_min_year():
    """Return a tuplet with most recent and oldest year folder available in './specs/'
    """
    years = [entry.name for entry in SPECS_PATH.iterdir() if entry.is_dir()]
    return (int(max(years, key=int)), int(min(years, key=int)))


MOST_RECENT_YEAR = _get_max_min_year()[0]
OLDEST_YEAR = _get_max_min_year()[1]

MODULES = ["ecd", "ecf", "efd_icms_ipi", "efd_pis_cofins"]

# We chose to hard-code the modules fields headers because it is easier to actualize
# them manually when necessary than mantaining a good heuristic algorithm catching these
# headers from raw CSV files.
# To define these headers manually, please use the script :
# `python -m spedextractor.get_mod_headers`
# which displays all the possible headers for each module.
MODULE_HEADER = {
    "ecd": [
        ("Nº", "index"),
        ("Campo", "code"),
        ("Descrição", "desc"),
        # key "type" reserved for the interpreted field dictionary
        ("Tipo", "spec_type"),
        ("Tamanho", "length"),
        ("Decimal", "decimal"),
        ("Valores Válidos", "spec_values"),
        # key "required" reserved for the interpreted field dictionary
        ("Obrigatório", "spec_required"),
        ("Regras de Validação do Campo", "rules"),
    ],
    "ecf": [
        ("Nº", "index"),
        ("Campo", "code"),
        ("Descrição", "desc"),
        ("Tipo", "spec_type"),
        ("Tamanho", "length"),
        ("Decimal", "decimal"),
        ("Valores Válidos", "spec_values"),
        ("Obrigatório", "spec_required"),
    ],
    "efd_icms_ipi": [
        ("Nº", "index"),
        ("Campo", "code"),
        ("Descrição", "desc"),
        ("Tipo", "spec_type"),
        ("Tam", "length"),
        ("Dec", "decimal"),
        ("Obrig", "spec_required"),
        ("Entr", "spec_in"),
        ("Saídas", "spec_out"),
    ],
    "efd_pis_cofins": [
        ("Nº", "index"),
        ("Campo", "code"),
        ("Descrição", "desc"),
        ("Tipo", "spec_type"),
        ("Tam", "length"),
        ("Dec", "decimal"),
        ("Obrig", "spec_required"),
    ],
}



#!/usr/bin/env python3

import logging

import camelot
import click
from PyPDF2 import PdfFileReader

from . import download
from .constants import MODULES, MOST_RECENT_YEAR, OLDEST_YEAR, SPECS_PATH

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)

START = 0
STEP = 10


def _limit_pages(pdf, limit=False):
    """Return the pages number to be extracted"""
    pdf_file = PdfFileReader(str(pdf.resolve()))
    return limit if limit else pdf_file.getNumPages()


def extract_mod_tables(mod, year, pdf=None, limit=False):
    if not pdf:
        download.download_mod_pdf(mod, year)
        pdf = SPECS_PATH / f"{year}" / "pdf" / f"{mod}.pdf"

    pdf_path = str(pdf.resolve())
    limit_pages = _limit_pages(pdf, limit)
    export_csv = SPECS_PATH / f"{year}" / f"{mod}" / "raw_camelot_csv"
    export_csv_path = str((export_csv / f"{mod}.csv").resolve())

    logger.info(
        f"> Extracting pdf {mod.upper()} {year} - {limit_pages} pages. "
        "It can take easily 5 minutes..."
    )
    # Creating directory if not existing
    export_csv.mkdir(parents=True, exist_ok=True)
    # Deleting previous extracted files if existing
    for file in export_csv.iterdir():
        file.unlink()

    # Extracting with camelot : we process the pages 10 by 10 to avoid malloc errors
    i = START
    while i < limit_pages:
        limit = min(i + STEP, limit_pages)
        logger.info(f"    pages {i} to {limit}...")
        tables = camelot.read_pdf(pdf_path, pages=f"{i}-{limit}", line_scale=40)
        tables.export(export_csv_path, f="csv", compress=False)
        i += STEP


@click.command()
@click.option(
    "--year",
    default=MOST_RECENT_YEAR,
    show_default=True,
    type=click.IntRange(OLDEST_YEAR, MOST_RECENT_YEAR),
    help="Operate on a specific year's folder, "
    f"can be between {OLDEST_YEAR} and {MOST_RECENT_YEAR}",
)
@click.option(
    "-l",
    "--limit",
    default=False,
    show_default=True,
    type=int,
    help="Extract a limited number of pdf pages",
)
def main(year, limit):
    """Extract tables from SPED modules pdf (from a given year) using camelot.

    The pdf must be present at './specs/YEAR/' and the extracted CSV files will be
    placed at './specs/YEAR/MODULE/raw_camelot_csv/'.

    If an option --limit is given, only the first pages will be parsed until the limit
    number."""
    logger.info(
        f"Extracting tables from the {year} SPED pdf. It can take a while "
        "(easily 20 minutes)"
    )
    for mod in MODULES:
        extract_mod_tables(mod, year, limit)


if __name__ == "__main__":
    main()



import pathlib

from setuptools import setup
from spedextractor import __version__

REPO = pathlib.Path(__file__).parent

README = (REPO / "README.rst").read_text()

setup(
    name="sped-extractor",
    version=__version__,
    description=(
        "Extrai e interpreta os registros e os campos das tabelas dos manuais do SPED "
        "(Sistema Público de Escrituração Digital), para os módulos "
        "ECD, ECF, EFD Contribuições (PIS, COFINS) e EFD ICMS IPI."
    ),
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://github.com/akretion/sped-extractor",
    author="Akretion",
    author_email="contact@akretion.com.br",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["spedextractor"],
    include_package_data=True,
    install_requires=[r.strip() for r in open("requirements.txt").read().splitlines()],
    setup_requires=["setuptools_scm"],
)
